import time
from nimbusinator import Nimbus, Command

if __name__ == '__main__':
    nim = Nimbus()
    cmd = Command(nim)
    nim.boot(skip_welcome_screen=True)
    cmd.set_mode(40)
    cmd.set_paper(9)
    cmd.cls()
    title = 'Nimbusinator'
    x = 10
    y = 50
    for d in [-1, 0, 1]:
        cmd.plot(title, (x+d, y-d), size=3, brush=0)
        cmd.plot(title, (x-d, y+d), size=3, brush=0)
        cmd.plot(title, (x, y-d), size=3, brush=0)
        cmd.plot(title, (x, y+d), size=3, brush=0)
        cmd.plot(title, (x+d, y), size=3, brush=0)
        cmd.plot(title, (x-d, y), size=3, brush=0)
    cmd.plot(title, (x, y), size=3, brush=15)
    time.sleep(10)
    nim.shutdown()


# RM Nimbus colour tables.  

# The Nimbus provided 16 colours in low-resolution mode.  In high
# resolution mode only 4 colours were available, although you can
# assign any of the 16 colours to the 4 slots in the palette.

# The RGB colour table
colour_table = {
    #     R    G    B
    0:  (  0,   0,   0),    # black
    1:  (  0,   0, 170),    # dark blue
    2:  (170,   0,   0),    # dark red
    3:  (170,   0, 170),    # purple
    4:  (  0, 170,   0),    # dark green
    5:  (  0, 170, 170),    # dark cyan
    6:  (170,  84,   0),    # brown
    7:  (170, 170, 170),    # light grey
    8:  ( 84,  84,  84),    # dark grey
    9:  ( 84,  84, 255),    # light blue
    10: (255,  84,  84),    # light red
    11: (255,  84, 255),    # light purple
    12: ( 84, 255,  84),    # light green
    13: ( 84, 255, 255),    # light cyan
    14: (255, 255,  84),    # yellow
    15: (255, 255, 255)     # white
}

# Default colours.  These are the default colour assignments for
# hi and lo resolution modes.  
default_colours = {
    'lo': {
        0:  0,    # black
        1:  1,    # dark blue
        2:  2,    # dark red
        3:  3,    # purple
        4:  4,    # dark green
        5:  5,    # dark cyan
        6:  6,    # brown
        7:  7,    # light grey
        8:  8,    # dark grey
        9:  9,    # light blue
        10: 10,   # light red
        11: 11,   # light purple
        12: 12,   # light green
        13: 13,   # light cyan
        14: 14,   # yellow
        15: 15    # white
    },
    'hi': {
        0: 1,     # dark blue
        1: 4,     # dark green
        2: 10,    # light red
        3: 15     # white
    }
}




from .tools import is_valid_colour, colrows_to_xy, recolour, fix_coord, points_to_nparray
from PIL import Image, ImageDraw, ImageOps
import numpy as np
import copy
import math
import os



class Command:
    """Nimbus commands

    This class contains the commands to control the Nimbus display.  The commands
    mimick the syntax found in RM Basic but re-written Pythonically.  In creating an 
    object of this class it must be bound to a pre-existing Nimbus object.  It is
    highly recommended to read the original Nimbus manuals (particularly RM Basic)
    to get a deeper understanding of how these commands originally worked.

    Args:
        nimbus (Nimbus): The Nimbus object to bind to

    """


    def __init__(self, nimbus):
        """Create a new Command object

        In creating an object of this class it must be bound to
        a pre-existing Nimbus object.

        Args:
            nimbus (Nimbus): The Nimbus object to bind to
        
        """

        # Validate params
        # To avoid circular import we can't put Nimbus in isintance but we can
        # assert that the object contains a string called title.
        #assert isinstance(nimbus.title, str), "Command object needs to be bound to a Nimbus object"
        self.nimbus = nimbus


    def set_paper(self, colour):
        """Set the paper colour
        
        Args:
            colour (int): Colour value (High-resolution: 0-3, low-resolution: 0-15)

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(colour, int), "The value of colour must be an integer, not {}".format(type(colour))
        assert is_valid_colour(self.nimbus, colour), "Colour {} is out-of-range for this screen mode".format(colour)

        self.nimbus.paper_colour = colour
        
    
    def set_colour(self, colour1, colour2):
        """Set a colour to a new colour

        Args:
            colour1 (int): The colour code to be changed
            colour2 (int): The new colour to be assigned to colour1
        
        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(colour1, int), "The value of colour1 must be an integer, not {}".format(type(colour1))
        assert isinstance(colour2, int), "The value of colour2 must be an integer, not {}".format(type(colour2))
        assert (colour2 >= 0 and colour2 <= 15), "The value of colour 2 must be >= 0 and <= 15, not {}".format(colour2)
        assert is_valid_colour(self.nimbus, colour1), "Colour1 {} is out-of-range for this screen mode".format(colour1)

        self.nimbus.runtime_colours[self.nimbus.screen_mode][colour1] = self.nimbus.DEFAULT_COLOURS['lo'][colour2]


    def set_border(self, colour):
        """Set the border colour
        
        Args:
            colour (int): Colour value (High-resolution: 0-3, low-resolution: 0-15)
            
        """
 
        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(colour, int), "The value of colour must be an integer, not {}".format(type(colour))
        assert is_valid_colour(self.nimbus, colour), "Colour {} is out-of-range for this screen mode".format(colour)

        self.nimbus.border_colour = colour
    

    def set_brush(self, colour):
        """Set the brush colour
        
        Args:
            colour (int): Colour value (High-resolution: 0-3, low-resolution: 0-15)
            
        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(colour, int), "The value of colour must be an integer, not {}".format(type(colour))
        assert is_valid_colour(self.nimbus, colour), "Colour {} is out-of-range for this screen mode".format(colour)

        self.nimbus.brush_colour = colour


    def set_pen(self, colour):
        """Set the pen colour

        Args:
            colour (int): Colour value (High-resolution: 0-3, low-resolution: 0-15)
            
        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(colour, int), "The value of colour must be an integer, not {}".format(type(colour))
        assert is_valid_colour(self.nimbus, colour), "Colour {} is out-of-range for this screen mode".format(colour)

        self.nimbus.pen_colour = colour


    def set_charset(self, charset):
        """Set the charset for text

        Args:
            charset (int): 0 is the standard font, 1 is the other font!

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(charset, int), "The value of charset must be an integer, not {}".format(type(charset))
        assert (charset == 0 or charset == 1), "The value of charset can be either 0 or 1, not {}".format(charset)

        self.nimbus.charset = charset


    def set_plot_font(self, plot_font):
        """Set the plot font

        Args:
            plot_font (int): 0 is the standard font, 1 is the other font!

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(plot_font, int), "The value of plot_font must be an integer, not {}".format(type(plot_font))
        assert (plot_font == 0 or plot_font == 1), "The value of plot_font can be either 0 or 1, not {}".format(plot_font)

        self.nimbus.plot_font = plot_font


    def ask_charset(self):
        """Return the current charset for text

        0 is the standard font, 1 is the other font

        Returns:
            int

        """

        return self.nimbus.charset


    def ask_plot_font(self):
        """Return the current plot_font

        0 is the standard font, 1 is the other font

        Returns:
            int

        """

        return self.nimbus.plot_font
        

    def ask_pen(self):
        """Return the current pen colour

        Returns:
            int

        """

        return self.nimbus.pen_colour


    def ask_paper(self):
        """Return the current paper colour

        Returns:
            int

        """

        return self.nimbus.paper_colour


    def ask_brush(self):
        """Return the current brush colour

        Returns:
            int
        """

        return self.nimbus.brush_colour


    def set_curpos(self, cursor_position):
        """Set the cursor position

        Args:
            cursor_position (tuple): The new cursor position (column, row)

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # Validate params
        assert isinstance(cursor_position, tuple), "The value of cursor_position must be a tuple, not {}".format(type(cursor_position))
        assert len(cursor_position) == 2, "The cursor_position tuple must have 2 values, not {}".format(len(cursor_position))
        for i in range(0, 2):
            assert isinstance(cursor_position[i], int), "This value in cursor_position must be an integer, not{}".format(type(cursor_position[i]))
        # Validate that cursor will still be on screen
        assert cursor_position[0] >= 0, "Negative column value in {} is not permitted".format(cursor_position)
        assert cursor_position[1] >= 0, "Negative row value in {} is not permitted".format(cursor_position)
        assert cursor_position[1] <= 25, "There are only 25 rows, not {}".format(cursor_position[1])
        if self.nimbus.screen_mode == 'hi':
            assert cursor_position[0] <= 80, "There are only 80 columns in this mode, not {}".format(cursor_position[0])
        if self.nimbus.screen_mode == 'lo':
            assert cursor_position[0] <= 40, "There are only 40 columns in this mode, not {}".format(cursor_position[0])

        # Now update the cursor position
        self.nimbus.cursor_position = cursor_position


    def ask_curpos(self):
        """Gets the current cursor position as column index, row index

        Returns:
            (int, int)

        """

        # Return cursor position
        return self.nimbus.cursor_position


    def ask_points_style(self):
        """Gets the current points style

        Returns:
            int

        """

        # Return points style
        return self.nimbus.points_style


    def cls(self):
        """Clear the screen of all text and graphics and reset cursor position

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # Wipe paper image in the Nimbus and reset cursor position
        self.nimbus.paper_image = self.nimbus.empty_paper()
        self.set_curpos((1, 1))


    def set_mode(self, columns):
        """Select either high-resolution or low-resolution screen mode

        In RM Basic the screen resolution was set by the number of columns:  40 for low-resolution and 80 for 
        high-resolution.  Any other values had no effect.  Nimbusinator is more strict and will yield an error
        if any other values are entered.  Check the original RM Basic manual for a description of how screen 
        resolutions worked on the Nimbus.

        MODE 40: 40 columns, 25 rows; 320 pixels wide, 250 pixels high; 16 colours

        MODE 80: 80 columns, 25 rows; 640 pixels wide, 250 pixels high (but doubled along vertical axis); 4 colours

        Args:
            columns (int): The number of colums (40 or 80)

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(columns, int), "The value of columns must be integer, not {}".format(columns)
        assert (columns == 80 or columns == 40), "The value of columns can be 80 or 40, not {}".format(columns)

        # restore default colors
        self.nimbus.runtime_colours = copy.deepcopy(self.nimbus.DEFAULT_COLOURS)

        # change screen size and set default colours accordingly
        if columns == 80:
            self.nimbus.screen_mode = 'hi'
            self.set_paper(0)
            self.set_border(0)
            self.set_brush(3)
            self.set_pen(3)
            self.set_curpos((1, 1))
            self.cls()
            return
        if columns == 40:
            self.nimbus.screen_mode = 'lo'
            self.set_paper(0)
            self.set_border(0)
            self.set_brush(15)
            self.set_pen(15)
            self.set_curpos((1, 1))
            self.cls()
            return


    def plonk_logo(self, coord):
        """Plonk the RM Nimbus logo on screen

        Args:
            coord (tuple): The (x, y) position to plonk the logo

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(coord, tuple), "The value of coord must be a tuple, not {}".format(coord)
        assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
        for i in range(0, 2):
            assert isinstance(coord[i], int), "This value in coord must be an integer, not{}".format(coord[i])

        self.nimbus.plonk_image_on_paper(self.nimbus.NIMBUS_LOGO, coord)


    def put(self, ascii_data):
        """Put a single character or string at the current cursor position

        Args:
            ascii_data (int/str): If an int is passed the corresponding ASCII character
                                    will be plotted.  If a string is passed then the 
                                    string will be printed without a terminating carriage
                                    return.

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # Validate params
        assert isinstance(ascii_data, (str, int)), "The value of ascii_data must be an integer or string, not {}".format(type(ascii_data))

        # Handle integer
        if isinstance(ascii_data, int):
            # validate in extended ASCII range
            assert (ascii_data >= 0 and ascii_data <= 255), "The value {} of ascii_data is outside the range of Extended ASCII (0-255)".format(ascii_data)
            # convert to char
            ascii_list = [chr(ascii_data)]
        # Handle string
        if isinstance(ascii_data, str):
            ascii_list = ascii_data
        
        # Put char or chars
        for ascii in ascii_list:

            # return if shutdown detected
            if not self.nimbus.running:
                return

            # If out of extended ASCII range replace with space
            if ord(ascii) > 255:
                # It's out of range
                ascii = ' '
            # Get char img
            char_img = self.nimbus.FONT_IMAGES[self.nimbus.charset][ord(ascii)]
            # Get screen position in pixels from cursor position
            curpos_xy = colrows_to_xy(self.nimbus.SCREEN_MODES[self.nimbus.screen_mode], self.nimbus.cursor_position)
            # Plot char and apply paper colour underneath char
            empty_char_image = recolour(self.nimbus, self.nimbus.EMPTY_CHAR_IMAGE, (0, 0, 0), self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][self.nimbus.paper_colour]], has_alpha=True)
            # Tweak position of paper background
            xp, yp = curpos_xy
            yp += 1
            paperpos_xy = (xp, yp)
            self.nimbus.plonk_image_on_paper(empty_char_image, paperpos_xy, transparent=True)
            # Overlay char, colourise and preserve paper colour
            char_img = recolour(self.nimbus, char_img, (0, 0, 0), self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][self.nimbus.pen_colour]], has_alpha=True)
            self.nimbus.plonk_image_on_paper(char_img, curpos_xy, transparent=True)
            # calculate new curpos, if over the right-hand side do carriage return
            new_column = self.nimbus.cursor_position[0] + 1
            if (self.nimbus.screen_mode == 'lo' and new_column > 40) or (self.nimbus.screen_mode == 'hi' and new_column > 80):
                # do carriage return
                new_column = 1  # return to left-hand side
                new_row = self.nimbus.cursor_position[1] + 1  # move down
                # if we're below, then screen, move screen data up 10 pixels and set
                # cursor to bottom of screen
                if new_row > 25:
                    new_row = 25
                    # Make a blank paper image and paste the old paper image 10 pixels higher 
                    # than the top and update actual paper image
                    new_paper_image = self.nimbus.empty_paper()
                    new_paper_image.paste(self.nimbus.paper_image, (0, -10))
                    self.nimbus.paper_image = new_paper_image
            else:
                # don't move cursor down
                new_row = self.nimbus.cursor_position[1]
            # move cursor
            self.set_curpos((new_column, new_row))


    def set_cursor(self, flag):
        """Show or hide cursor

        Args:
            flag (boolean): True to show cursor, False to hide

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(flag, bool), "The value of flag must be boolean, not {}".format(flag)

        self.nimbus.cursor_enabled = flag


    def print(self, text):
        """Print a string with carriage return at end

        Args:
            text (str): The text to be printed

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # Validate params
        assert isinstance(text, str), "The value of text must be a string, not {}".format(type(text))

        # Put the string and then maybe CR
        self.put(text)
        # Carriage return?
        col, row = self.ask_curpos()
        if col > 1:
            # Yep - smash the cursor off the screen and use put to force CR
            self.nimbus.cursor_position = (255, row)
            self.put('X')


    def flush(self):
        """Clears the keyboard buffer

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        self.nimbus.keyboard_buffer = []


    def gets(self):
        """Get the oldest char in the keyboard buffer

        Equivalent to GET$ in RM Basic except it will also return the following
        values if a control key is pressed: '__F1__', '__F2__', ... , '__F10__',
        '__CTRL_L__', '__CTRL_R__', '__ENTER__', '__BACKSPACE__', '__ESC__',
        '__DELETE__', '__HOME__', '__PAGE_UP__', '__END__', '__PAGE_DOWN__',
        '__INSERT__', '__UP__', '__DOWN__', '__LEFT__', '__RIGHT__', '__TAB__'

        Returns:
            str
        
        """

        # If the buffer isn't empty pop the last char
        # and return it, otherwise return empty str
        if len(self.nimbus.keyboard_buffer) > 0:
            popped_value = self.nimbus.keyboard_buffer.pop(0)
            # Handle control key or char
            # An integer with -1 indicates control char, so get
            # it from the control char variable, otherwise just
            # a char so return that
            if isinstance(popped_value, int):
                # Is probably a control char
                if popped_value == -1:
                    # It *is* a control char - return it
                    return self.nimbus.control_char
            else:
                # It's just a char
                return popped_value
        else:
            return ''        


    def input(self, prompt):
        """Collects keyboard input until user presses ENTER then returns the input as a string

        Args:
            prompt (str): Message to be printed

        Returns:
            str
        
        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # Validata params
        assert isinstance(prompt, str), "The value of prompt must be a string, not {}".format(type(prompt))

        # Get max columns for this screen mode
        if self.nimbus.screen_mode == 'lo':
            max_columns = 40
        if self.nimbus.screen_mode == 'hi':
            max_columns = 80
        # Flush buffer, reset enter + delete flag
        self.flush()
        self.nimbus.enter_was_pressed = False
        self.nimbus.backspace_was_pressed = False
        # Print the prompt and get start cursor position
        self.put(prompt)
        # Collect response in this string:
        response = ''
        # Collect and echo chars from buffer until enter was pressed
        while not self.nimbus.enter_was_pressed and self.nimbus.running:
            new_char = self.gets()
            # don't add control key hits:
            if new_char not in self.nimbus.SUPPORTED_CONTROL_KEYS.values():
                response += new_char
                self.put(new_char)
            # Handle delete
            if self.nimbus.backspace_was_pressed and len(response) == 0:
                self.nimbus.backspace_was_pressed = False
            if self.nimbus.backspace_was_pressed and len(response) > 0:
                now_col, now_row = self.ask_curpos()
                response = response[:-1]
                # Move cursor left
                # If we're about to move off the left-hand side of the screen
                # it must be because we're on a line below where the input
                # started.  So, we need to move up one row and locate cursor
                # at right-hand side, wipe whatever char is there, then re-
                # position
                next_col = now_col - 1
                if next_col == 0:
                    # Go up one row
                    next_col = max_columns
                    next_row = now_row - 1
                else:
                    # Go back one column
                    next_row = now_row
                # Wipe char in this location, reposition, and reset flag
                self.set_curpos((next_col, next_row))
                self.put(' ')
                self.set_curpos((next_col, next_row))
                self.nimbus.backspace_was_pressed = False
        # Enter was pressed, flush buffer and reset enter flag
        self.nimbus.enter_was_pressed = False
        self.flush()
        # Force carriage return by smashing the cursor off the screen
        col, row = self.ask_curpos()
        self.nimbus.cursor_position = (255, row)
        self.put('X')

        # Return string
        return response


    def plot(self, text, coord, size=1, brush=None, direction=0, font=None):
        """Plot text on the screen

        Args:
            text (str): The text to be plotted
            coord (tuple): The (x, y) position of the text
            size (int), optional: Font size. To elongate pass a tuple (x_size, y_size)
            brush (int), optional: Brush colour
            direction (int), optional: 0=normal, 1=-90deg, 2=180deg, 3=-270deg
            font (int), optional: 0 is the standard font, 1 is the other font

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # Validate params
        assert isinstance(text, str), "The value of text must be a string, not {}".format(type(str))
        assert isinstance(coord, tuple), "The value of coord must be a tuple, not {}".format(type(coord))
        assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
        for i in range(0, 2):
            assert isinstance(coord[i], int), "The values in coord {} must be integer, not {}".format(coord, type(coords[i]))      
        if size is not None:
            assert isinstance(size, (int, tuple)), "The value of size must be an integer or tuple, not {}".format(type(size))
        if isinstance(size, tuple):
            for i in range(0, 2):
                assert isinstance(size[i], int), "The values in size {} must be integer, not {}".format(size, type(size[i]))  
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        if brush is not None:
            assert is_valid_colour(self.nimbus, brush), "Brush colour {} is out-of-range for this screen mode".format(brush)
        assert isinstance(direction, int), "The value of direction must be an integer, not {}".format(type(direction))
        assert isinstance(font, (type(None), int)), "The value of font must be an integer, not {}".format(type(font))
        assert (font == 0 or font == 1 or font is None), "The value of font can be 0 or 1, not {}".format(font)

        # Handle brush colour
        if brush is None:
            brush = self.nimbus.brush_colour

        # Handle font
        if font is None:
            font = self.nimbus.plot_font

        # Create a temporary image of the plotted text
        plot_img_width = len(text) * 10
        plot_img = Image.new(
            'RGBA',
            (plot_img_width, 10), 
            (0, 0, 0, 0)   # 1
        )

        # Plot chars on plot_img
        x = 0
        for char in text:

            # return if shutdown detected
            if not self.nimbus.running:
                return

            # If out of extended ASCII range replace with space
            if ord(char) > 255:
                # It's out of range
                char = ' '
            # Get char image and recolour it
            char_img = self.nimbus.FONT_IMAGES[font][ord(char)]
            char_img = recolour(self.nimbus, char_img, (0, 0, 0), self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][brush]], has_alpha=True)
            # Plot char and increment x
            plot_img.paste(char_img, (x, 0), mask=char_img)
            x += 8
        
        # resize
        if isinstance(size, tuple):
            # tuple: extract x_size, y_size
            x_size, y_size = size
        else:
            x_size = size
            y_size = size
        new_size = (plot_img.size[0] * x_size, plot_img.size[1] * y_size)
        plot_img = plot_img.resize(new_size, resample=Image.NEAREST)
        
        # rotate
        for i in range(0, direction):
            plot_img = plot_img.transpose(Image.ROTATE_90)
        
        # rebuild screen and done
        self.nimbus.plonk_image_on_paper(plot_img, coord, transparent=True)


    def area(self, coord_list, brush=None, scale=1):
        """Draw a filled polygon

        Args:
            coord_list (list): A list of (x, y) tuples
            brush (int), optional: Colour value (High-resolution: 0-3, low-resolution: 0-15)
            scale (int), optional: Scale factor. To elongate pass a tuple (x_size, y_size)
        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(coord_list, list), "coord_list should contain a list, not {}".format(type(coord_list))
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        if brush is not None:
            assert is_valid_colour(self.nimbus, brush), "Brush colour {} is out-of-range for this screen mode".format(brush)
        for coord in coord_list:
            assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
            for i in range(0, 2):
                assert isinstance(coord[i], int), "The values in coord {} must be integer, not {}".format(coord, type(coord[i]))
        assert isinstance(scale, (int, tuple, float)), "The value of scale must be an integer, float or tuple, not {}".format(type(scale))
        if isinstance(scale, tuple):
            for i in range(0, 2):
                assert isinstance(scale[i], (int, float)), "The values in scale {} must be integer or float, not {}".format(scale, type(scale[i]))
                assert scale[i] > 0, "The value {} in scale must be greater than zero".format(scale[i])
        if isinstance(scale, (int, float)):
            assert scale > 0, "The value {} in scale must be greater than zero".format(scale)

        # if default brush value then get current brush colour
        if brush is None:
            brush = self.nimbus.brush_colour

        if isinstance(scale, tuple):
            # tuple: extract x_size, y_size
            x_size, y_size = scale
        else:
            x_size = scale
            y_size = scale

        # coordinate as the origin.
        origin = coord_list[0]
        x_org, y_org = origin

        for i in range(0, len(coord_list)):
            if coord_list[i] == origin:
                # don't scale origin
                x, y = coord_list[i]
                coord_list[i] = fix_coord(self.nimbus.SCREEN_MODES[self.nimbus.screen_mode], (x, y))
            else:
                x, y = coord_list[i]
                x += ((x - x_org) * (x_size - 1))
                y += ((y - y_org) * (y_size - 1))
                coord_list[i] =  fix_coord(self.nimbus.SCREEN_MODES[self.nimbus.screen_mode], (x, y))

        # draw area on temp image
        draw = ImageDraw.Draw(self.nimbus.paper_image)
        rgb = self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][brush]]
        draw.polygon(coord_list, outline=rgb, fill=rgb)


    def line(self, coord_list, brush=None):
        """Draw one or more connected straight lines

        Args:
            coord_list (list): A list of (x, y) tuples
            brush (int), optional: Colour value (High-resolution: 0-3, low-resolution: 0-15)

        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(coord_list, list), "coord_list should contain a list, not {}".format(type(coord_list))
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        if brush is not None:
            assert is_valid_colour(self.nimbus, brush), "Brush colour {} is out-of-range for this screen mode".format(brush)
        for coord in coord_list:
            assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
            for i in range(0, 2):
                assert isinstance(coord[i], int), "The values in coord {} must be integer, not {}".format(coord, type(coord[i])) 

        # if default brush value then get current brush colour
        if brush is None:
            brush = self.nimbus.brush_colour

        # correct coords
        for i in range(0, len(coord_list)):
            coord_list[i] = fix_coord(self.nimbus.SCREEN_MODES[self.nimbus.screen_mode], coord_list[i])

        # draw lines
        draw = ImageDraw.Draw(self.nimbus.paper_image)
        rgb = self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][brush]]
        draw.line(coord_list, fill=rgb)


    def circle(self, radius, coord_list, brush=None):
        """Draw one or more circles

        Args:
            radius (int): The radius of the circle
            coord_list (list): A list of (x, y) tuples
            brush (int), optional: Colour value (High-resolution: 0-3, low-resolution: 0-15)
        
        """

        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(radius, int), "radius must be an integer, not {}".format(type(radius))
        assert radius > 0, "radius must be greater than zero, not {}".format(radius)
        assert isinstance(coord_list, list), "coord_list should contain a list, not {}".format(type(coord_list))
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        if brush is not None:
            assert is_valid_colour(self.nimbus, brush), "Brush colour {} is out-of-range for this screen mode".format(brush)
        for coord in coord_list:
            assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
            for i in range(0, 2):
                assert isinstance(coord[i], int), "The values in coord {} must be integer, not {}".format(coord, type(coord[i]))

        # if default brush value then get current brush colour
        if brush is None:
            brush = self.nimbus.brush_colour

        # draw cirtles
        draw = ImageDraw.Draw(self.nimbus.paper_image)
        rgb = self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][brush]]
        for coord in coord_list:

            # return if shutdown detected
            if not self.nimbus.running:
                return

            # define 'bounding box'
            r = radius
            x, y = coord
            box = [(x-r, y-r), (x+r, y+r)]
            draw.ellipse(box, outline=rgb, fill=rgb)


    def slice_(self, radius, from_angle, to_angle, coord_list, brush=None):
        """Draw one or more pie slices

        Args:
            radius (int): The radius of the slice
            from_angle (int): The starting angle (degrees)
            to_angle (int): The finishing angle (degrees)
            coord_list (list): A list of (x, y) tuples
        
        """
        
        # return if shutdown detected
        if not self.nimbus.running:
            return

        # validate params
        assert isinstance(radius, int), "radius must be an integer, not {}".format(type(radius))
        assert radius > 0, "radius must be greater than zero, not {}".format(radius)
        assert isinstance(from_angle, int), "from_angle must be an integer, not {}".format(type(from_angle))
        assert isinstance(to_angle, int), "to_angle must be an integer, not {}".format(type(to_angle))
        assert from_angle >= 0 and from_angle <= 360, "from_angle must be between 0 and 360 degrees, not {}".format(from_angle)
        assert to_angle >=0 and to_angle <= 360, "to_angle must be between 0 and 360 degrees, not {}".format(to_angle)
        assert isinstance(coord_list, list), "coord_list should contain a list, not {}".format(type(coord_list))
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        if brush is not None:
            assert is_valid_colour(self.nimbus, brush), "Brush colour {} is out-of-range for this screen mode".format(brush)
        for coord in coord_list:
            assert len(coord) == 2, "The coord tuple must have 2 values, not {}".format(len(coord))
            for i in range(0, 2):
                assert isinstance(coord[i], int), "The values in coord {} must be integer, not {}".format(coord, type(coord[i]))

        # if default brush value then get current brush colour
        if brush is None:
            brush = self.nimbus.brush_colour

        # draw slices      
        draw = ImageDraw.Draw(self.nimbus.paper_image)
        rgb = self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][brush]]
        for coord in coord_list:

            # return if shutdown detected
            if not self.nimbus.running:
                return

            # define 'bounding box'
            r = radius
            x, y = coord
            box = [(x-r, y-r), (x+r, y+r)]
            draw.pieslice(box, from_angle, to_angle, outline=rgb, fill=rgb)


    def set_points_style(self, points_style, points_list):
        """Set a points style to that define in a points list

        You can define up to 256 points styles (0 - 255).  The points
        list is similar to the way patterns are define in RM Basic.
        Each row in the points style is define by a string, with full-stop
        chars (.) indicating a filled pixel, and spaces ( ) indicating a
        filled pixel.  The size of a points style is 8x8 pixels.

        Args:
            points_style (int): The points style number (0 255)
            points_list (list): The points list as defined above
        
        """

        # validate params
        assert isinstance(points_style, int), "points_style must be an integer, not {}".format(type(points_style))
        assert points_style >= 0 and points_style <= 255, "points_style must be in range 0-255, not {}".format(points_style)
        assert isinstance(points_list, list), "points_list must be a list, not {}".format(type(points_list))
        assert len(points_list) == 8, "there must be 8 items in points_list, not {}".format(len(points_list))
        for points in points_list:
            assert isinstance(points, str), "the item {} in points_list must be a string, not {}".format(points, type(points))
            assert len(points) == 8, "the length of this item must be 8, not {}".format(len(points))

        # create and assign the points style
        self.nimbus.points_styles[points_style] = points_to_nparray(points_list)

    
    def points(self, coord_list, brush=None, size=1, style=None):
        """Draw points on the screen

        Args:
            coord_list (list): 
            style (int): The points style number (0 255)
            coord_list (list): A list of (x, y) tuples
            brush (int), optional: Colour value (High-resolution: 0-3, low-resolution: 0-15)
            size (int), optional: Font size. To elongate pass a tuple (x_size, y_size)

        """

        assert isinstance(style, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(style))
        if style is not None:
            assert style >= 0 and style <= 255, "style must be in range 0-255, not {}".format(style)
        assert isinstance(size, (type(None), int)), "The value of size must be None or an integer, not {}".format(type(size))
        assert isinstance(brush, (type(None), int)), "The value of brush must be None or an integer, not {}".format(type(brush))
        if brush is not None:
            assert is_valid_colour(self.nimbus, brush), "Brush colour {} is out-of-range for this screen mode".format(brush)
        assert isinstance(size, (int, tuple)), "The value of size must be an integer or tuple, not {}".format(type(size))
        if isinstance(size, tuple):
            for i in range(0, 2):
                assert isinstance(size[i], int), "The values in size {} must be integer, not {}".format(size, type(size[i]))  

        # if default brush value then get current brush colour, same for the other optional params
        if brush is None:
            brush = self.nimbus.brush_colour
        if size is None:
            size = 1
        if style is None:
            style = self.nimbus.points_style

        # draw the points
        for coord in coord_list:
            # make a temp image
            temp_img = Image.fromarray(self.nimbus.points_styles[style], mode='RGBA')  
            # scale image
            if isinstance(size, tuple):
                # tuple: extract x_size, y_size
                x_size, y_size = size
            else:
                x_size = size
                y_size = size
            new_size = (temp_img.size[0] * x_size, temp_img.size[1] * y_size)
            temp_img = temp_img.resize(new_size, resample=Image.NEAREST)
            # draw it
            self.nimbus.plonk_image_on_paper(recolour(self.nimbus, temp_img, (0, 0, 0), self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][brush]], has_alpha=True), coord, transparent=True)


    def fetch(self, block_number, filename):
        """Fetch an image from disk and allocate it to an image block

        Equivalent to FETCH in the ANIMATE extension.  This is an experimental
        feature.  All major image formats are supported but in this loading 
        images with alpha layers may break.  It is also very slow!

        Args:
            block_number (int): The block number to store the image in
            filename (str): The filename of the image to fetch
        """          

        assert isinstance(block_number, int), "block_number must be an integer, not {}".format(type(block_number))
        assert block_number >=0 and block_number <= 255, "block_number must be in the range 0-255, not {}".format(block_number)
        assert isinstance(filename, str), "filename must be a string, not {}".format(type(filename))
        assert os.path.isfile(filename), "Image file {} not found".format(filename)


        def get_closest_nimbus_colour(possible_rgbs, rgb):
            """Find the closest Nimbus colour in the current runtime palette"""

            def distance(c1, c2):
                """Calculate the distance between two RGB colours"""
                (r1,g1,b1) = c1
                (r2,g2,b2) = c2
                return math.sqrt((r1 - r2)**2 + (g1 - g2) ** 2 + (b1 - b2) **2)


            nearest_match = {'rgb': (0, 0, 0), 'score': 1000}
            for i in range(0, len(possible_rgbs)):
                score = distance(rgb, possible_rgbs[i])
                if score < nearest_match['score']:
                    nearest_match = {'rgb': possible_rgbs[i], 'score': score}
            return nearest_match['rgb']

        # Open the image and convert to numpy array
        img = Image.open(filename)
        img = np.array(img)

        # get possible runtime rgbs
        possible_colours = self.nimbus.runtime_colours[self.nimbus.screen_mode]
        possible_rgbs = []
        for col in possible_colours:
            possible_rgbs.append(self.nimbus.COLOUR_TABLE[self.nimbus.runtime_colours[self.nimbus.screen_mode][col]])

        # Reassign all RGBs to nearest in runtime palette
        for y in range(0, img.shape[0]):
            for x in range(0, img.shape[1]):
                rgb_array = img[y][x]
                rgb_tuple = (rgb_array[0], rgb_array[1], rgb_array[2])
                nearest_nimbus_rgb_tuple = get_closest_nimbus_colour(possible_rgbs, rgb_tuple)
                img[y][x] = [ nearest_nimbus_rgb_tuple[0], nearest_nimbus_rgb_tuple[1], nearest_nimbus_rgb_tuple[2] ]

        # All done, assign the image to the required image block
        self.nimbus.image_blocks[block_number] = Image.fromarray(img, mode='RGB')
        

    def writeblock(self, block_number, coord):
        """Write an image block to the screen

        Equivalent to WRITEBLOCK in the ANIMATE extension.  Alpha layers are not
        yet supported - see documentation for fetch command.

        Args:
            block_number (int): The block number of the image block
            coord (tuple): The (x, y) position of the image
        
        """

        assert isinstance(block_number, int), "block_number must be an integer, not {}".format(type(block_number))
        assert block_number >=0 and block_number <= 255, "block_number must be in the range 0-255, not {}".format(block_number)
        assert self.nimbus.image_blocks[block_number] is not None, "block_number {} is empty".format(block_number)

        # Draw the memory block
        self.nimbus.plonk_image_on_paper(self.nimbus.image_blocks[block_number], coord, transparent=False)


import os
import sys
import time
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
import pygame
import copy
import numpy as np
from .tools import logo, message, pil_to_pygame_image, fix_coord, char_image_selecta, colrows_to_xy, recolour, points_to_nparray
from .colour_table import colour_table, default_colours
from PIL import Image, ImageDraw, ImageColor
import threading
import random
from pynput import keyboard
import simpleaudio as sa
from .welcome import welcome
from .command import Command
from .supported_control_keys import supported_control_keys


# get full path of this script
real_path = os.path.dirname(os.path.realpath(__file__))


class Nimbus:
    """Nimbus video display class.

    This class represents the Nimbus video display that will host the user
    interface for your application.  When created, the new Nimbus object 
    will not be visible until the boot() method has been called.

    Args:
        full_screen (bool), optional: Full screen mode
        title (str), optional: The title of the display window

    """

    def __init__(self, full_screen=False, title='Nimbusinator', border_size=40):
        """Create a new Nimbus object

        When created, the new Nimbus object  will
        not be visible until the boot() method has been called.

        Args:
            full_screen (bool), optional: Full screen mode
            title (str), optional: The title of the display window

        """

        print(logo)

        # Constants
        self.SCREEN_MODES = {                           # The absolute screen resolution for
            'hi': (640, 250),                           # high (80 column) and low-res (40
            'lo': (320, 250)                            # column) modes
            }
        self.FONT_IMAGES = self.__load_fonts()              # Font images
        
        test = np.array(self.FONT_IMAGES[0][70])
        for row in test:
            out = ''
            for col in row:
                colstr = str(col)
                if colstr == '[  0   0   0 255]':
                    char = 'X'
                else:
                    char = '_'
                out += '{}'.format(char)


        self.SUPPORTED_CONTROL_KEYS = supported_control_keys
        self.COLOUR_TABLE = colour_table                # This is the RGB colour table from the Nimbus
        self.DEFAULT_COLOURS = default_colours          # The default colours which cannot be changed during runtime
        self.NORMALIZED_PAPER_SIZE = (640, 500)         # After drawing graphics and chars the paper is normalized to this size
        # Background size is the normalized paper size plus 2x border size:
        background_width = self.NORMALIZED_PAPER_SIZE[0] + (2 * border_size)
        background_height = self.NORMALIZED_PAPER_SIZE[1] + (2 * border_size)
        self.BACKGROUND_SIZE = (background_width, background_height)
        # RM Nimbus logo
        logo_path = os.path.join(real_path, 'data', 'rm-nimbus-logo.bmp')
        self.NIMBUS_LOGO = Image.open(logo_path)
        self.CURSOR_IMAGE = Image.new(
            'RGB',
            (10, 2), 
            (0, 0, 0)
        )
        self.EMPTY_CHAR_IMAGE = Image.new(
            'RGB',
            (10, 10),
            (0, 0, 0)
        )

        # Settings
        self.full_screen = full_screen
        self.title = title
        self.border_size = border_size

        # Variables
        self.screen_mode = 'hi'
        self.runtime_colours = copy.deepcopy(self.DEFAULT_COLOURS)   # These colours can be changed at runtime
        self.paper_colour = 1
        self.border_colour = 1
        self.brush_colour = 3
        self.pen_colour = 3
        self.charset = 0
        self.plot_font = 0
        self.cursor_position = (1, 1)
        self.cursor_enabled = False
        self.keyboard_buffer = []
        # Set up default and user-define points
        self.points_styles = self.__init_points_styles()
        self.points_style = 0
        self.control_char = ''
        self.image_blocks = [None] * 255

        # Initialize with empty paper
        self.paper_image = self.empty_paper()

        # Status flags
        self.running = False                          # Set to True to run the Nimbus and the display
        self.__cursor_flash = False                     # Make cursor visible if True
        self.ctrl_pressed = False
        self.enter_was_pressed = False
        self.backspace_was_pressed = False
        self.floppy_is_running = False


    def __init_points_styles(self):
        """Initialize a list of default points styles and empty styles"""

        styles = []
        for i in range(0, 256):
            styles.append(
                Image.new(
                    'RGBA',
                    (8, 8),
                    (255, 255, 255, 0)
                )       
            )
        return styles


    def __load_fonts(self):
        """Load fonts from PNG files

        Returns:
            (dict): A dict of PIL images for font 0 and font 1

        """

        fonts = {}
        for font in range(0, 2):
            font_img_path = os.path.join(real_path, 'data', 'font{}-alpha.png'.format(font))
            font_img = Image.open(font_img_path)
            fonts[font] = []
            for ascii_code in range(0, 256):
                fonts[font].append(char_image_selecta(font_img, ascii_code, font))
        return fonts


    def empty_paper(self):
        """Return empty paper filled with the current paper colour"""

        return Image.new(
            'RGB',
            self.SCREEN_MODES[self.screen_mode], 
            self.__get_rgb(self.paper_colour)
        )


    def __get_rgb(self, colour):
        """Get an RGB tuple for a given colour code and current screen mode

        Args:
            colours (int): The Nimbus colour
        
        Returns:
            (tuple): The RGB tuple

        """

        return self.COLOUR_TABLE[self.runtime_colours[self.screen_mode][colour]]


    def plonk_image_on_paper(self, img, coord, transparent=False):
        """Plonk a PIL image somewhere on the paper
        
        The passed coordinates are consistent with RM Nimbus implemented, i.e.
        bottom-left of screen = (0, 0) and any images are located relative to
        their bottom-left corner

        Args:
            img (PIL image): The image to be plonked
            coord (tuple): The coordinate tuple (x, y)
            transparent (bool): True if image contains an alpha layer for transparency

        """

        # Flip y
        x, y = fix_coord(self.SCREEN_MODES[self.screen_mode], coord)

        # And change point-of-reference to bottom-left corner position
        y -= img.size[1]
        
        # Then paste image
        if transparent:
            self.paper_image.paste(img, (x, y), mask=img)
        else:
            self.paper_image.paste(img, (x, y))


    def __cycle_cursor_flash(self):
        """Cycle the cursor flash

        This needs to run in it's own thread

        """
        while self.running:
            # If the Nimbus is running flip cursor flash every 0.5 secs
            time.sleep(0.5)
            cursor_flash = self.__cursor_flash
            self.__cursor_flash = not cursor_flash
        return


    def __display_loop(self):
        """The display loop

        Normalize paper image.  Overlay normalized image on background
        image to make the video image.  If not in full screen mode
        just show the video image.  If in full screen mode scale the
        video image to the display height, overlay it on a background
        image that has the same dimensions as the display, and show.
        The loop breaks if self.running becomes false.  This method
        must be run in its own thread.

        """

        while self.running:
            # Overlay flash cursor if enabled
            paper_image_plus_cursor = self.paper_image.copy()
            if self.__cursor_flash and self.cursor_enabled:
                # Recolour cursor image to pen colour
                cursor_image = recolour(self, self.CURSOR_IMAGE, (0, 0, 0), self.COLOUR_TABLE[self.runtime_colours[self.screen_mode][self.pen_colour]], has_alpha=True)
                # Paste cursor image on paper_image_plus_cursor at x, y
                # Flip y
                x, y = fix_coord(self.SCREEN_MODES[self.screen_mode], colrows_to_xy(self.SCREEN_MODES[self.screen_mode], self.cursor_position))
                # Change point-of-reference to bottom-left corner position
                y -= 2
                # And paste cursor
                paper_image_plus_cursor.paste(cursor_image, (x, y))
            # Normalized paper image
            normalized_paper_image = paper_image_plus_cursor.resize(self.NORMALIZED_PAPER_SIZE, resample=Image.NEAREST)
            # Create background image with border colour
            background_image = Image.new(
                'RGB',
                self.BACKGROUND_SIZE,
                self.__get_rgb(self.border_colour)
            )
            # Overlay normalized paper image on background
            background_image.paste(normalized_paper_image, (self.border_size, self.border_size))
            # Handle full screen
            if self.full_screen:
                # Calculate new dimensions, resize and calculate x offset
                scale = self.__full_screen_display_size[1] / self.BACKGROUND_SIZE[1]
                new_size = (int(self.BACKGROUND_SIZE[0] * scale), self.__full_screen_display_size[1])
                background_image = background_image.resize(new_size, resample=Image.BICUBIC)
                display_x_offset = int((self.__full_screen_display_size[0] - new_size[0]) / 2)
            else:
                # Windows, so no x offset or resizing
                display_x_offset = 0
            # Create the video image, blit it and flip it
            video_image = pil_to_pygame_image(background_image)
            self.__pygame_display.blit(video_image, (display_x_offset, 0))
            pygame.display.flip()


    def __on_key_press(self, key):
        """Handle control key presses

        """
        
        # Printable chars go straight into the buffer
        try:
            self.keyboard_buffer.append(key.char)
            # BUT - if CTRL-C situation then shutdown!
            if self.ctrl_pressed and key.char.lower() == 'c':
                message('CTRL-C detected')
                self.shutdown()
        except AttributeError:
            # Handle CTRL released
            if key == keyboard.Key.ctrl_l or keyboard.Key.ctrl_r:
                self.ctrl_pressed = True
            # Handle ENTER hit
            if key == keyboard.Key.enter:
                self.enter_was_pressed = True
            # Handle BACKSPACE hit
            if key == keyboard.Key.backspace:
                self.backspace_was_pressed = True
            # Also add spaces to buffer
            if key == keyboard.Key.space:
                self.keyboard_buffer.append(' ')
            # The following keys can be detected by gets by sending -1 to the 
            # buffer and logging the key in control_char
            for control_key, control_key_string in self.SUPPORTED_CONTROL_KEYS.items():
                if key == control_key:
                    self.keyboard_buffer.append(-1)
                    self.control_char = control_key_string
                    break



    def __on_key_release(self, key):
        """Handle control key releases

        """

        # Handle CTRL released
        if key == keyboard.Key.ctrl_l or keyboard.Key.ctrl_r:
            self.ctrl_pressed = False


    def __floppy_drive_effects(self):
        """Simulate floppy drive grinding

        This has to be run a thread.  Two drive sounds are supplied: a
        short 'dot' sounds, and a long 'dash' sound, with 4 variations
        of each.  Semi-random patterns of dashes and dots are generated 
        with randomly-selected variations while the floppy_is_running
        flag is True.

        """
        while self.running:
            # Load sound effects
            dash = []
            dot = []
            for i in range(1, 5):
                # Abort if shutting down
                if not self.running:
                    return
                # Otherwise load sound effects
                dash_path = os.path.join(real_path, 'data', 'floppy-dash{}.wav'.format(i))
                dot_path = os.path.join(real_path, 'data', 'floppy-dot{}.wav'.format(i))
                dash.append(sa.WaveObject.from_wave_file(dash_path))
                dot.append(sa.WaveObject.from_wave_file(dot_path))
            # If flag is True play a bunch of grinding floppy drive sounds
            if self.floppy_is_running and self.running:
                # play dash (pick one at random)
                play_obj = dash[random.randint(0, 3)].play()
                play_obj.wait_done()
                # Abort is shutting down
                if not self.running:
                    return
                # Otherwise play random number of dots
                for i in range(0, random.randint(1, 5)):
                    # break out if drive is stopped or shutting down
                    if not self.floppy_is_running or not self.running:
                        break
                    else:
                        # pick one at random
                        play_obj = dot[random.randint(0, 3)].play()
                        play_obj.wait_done()


    def run_floppy(self, flag):
        """Run or stop the floppy drive sound effects

        Augment your user experience with the industrial melodies of a 
        1980s PC floppy drive

        Args:
            flag (boolean): True to run the drive, False to stop

        """

        self.floppy_is_running = flag


    def boot(self, skip_welcome_screen=False):
        """Boot the Nimbus
        
        This will reveal the display screen and start all the Nimbus-related processed.  By 
        default the Nimbus will go through the Welcome Screen and a mock boot-up sequence 
        before returning control to your app.  You can bypass the Welcome Screen by passing 
        skip_welcome_screen=True.  Once boot() has been called, the Nimbus can be interrupted 
        and stopped programmatically by calling the shutdown() method, or during runtime by 
        the user pressing CTRL-C. 

        Args:
            skip_welcome_screen (bool), optional: Bypass the Welcome Screen and boot sequence
            
        """

        # Don't boot if already running
        if self.running:
            message('The Nimbus is already running')
            return

        message('Booting up')

        # Initialize pygame and handle full screen
        pygame.init()
        pygame.display.set_caption(self.title)
        if self.full_screen:
            # Full screen - get full screen size as display size and set pygame flags
            self.__full_screen_display_size = (pygame.display.Info().current_w, pygame.display.Info().current_h)
            display_size = self.__full_screen_display_size
            flags = pygame.FULLSCREEN
        else:
            # Run in window - use background size as display size and set pygame flags
            display_size = self.BACKGROUND_SIZE
            flags = 0
        self.__pygame_display = pygame.display.set_mode(display_size, flags=flags)

        # Set flags
        self.running = True

        # Start display loop in a thread
        self.__display_thread = threading.Thread(target=self.__display_loop, args=())
        self.__display_thread.start()
        # Fire up cursor in another thread
        self.__cursor_thread = threading.Thread(target=self.__cycle_cursor_flash, args=())
        self.__cursor_thread.start()
        # Fire up the floppy disk effects in another
        self.__floppy_thread = threading.Thread(target=self.__floppy_drive_effects, args=())
        self.__floppy_thread.start()
        # Fire up the keyboard listener
        self.__keyboard_listener = keyboard.Listener(
            on_press=self.__on_key_press,
            on_release=self.__on_key_release)
        self.__keyboard_listener.start()

        if skip_welcome_screen:
            # don't bother with welcome screen
            Command(self).set_mode(80)
            message('Done')
            return
        else:
            # roll the welcome screen
            welcome(Command(self), self)
            Command(self).set_mode(80)
            message('Done')
    
    def shutdown(self):
        """Shut down the Nimbus
        
        Stops all commands from executing, stops all Nimbus-related processes,
        closes the display window and halts program execution.

        """

        # Passshutdown if not running
        if not self.running:
            return

        message('Shutting down')

        # Set flags
        self.running = False

        # Join threads
        self.__display_thread.join()
        self.__cursor_thread.join()

        # Stop keyboard listener
        self.__keyboard_listener.stop()

        # Quit pygame and exit
        pygame.quit()
        sys.exit(1)


    def sleep(self, sleep_time):
        """Pause execution like time.sleep()

        Unlike time.sleep(), this built-in sleep method will be interrupted
        if the user hits CTRL-C.  Sleep time measured in seconds.

        """

        # Validate params
        assert isinstance(sleep_time, (int, float)), "The value of sleep_time must be an integer or float, not {}".format(type(sleep_time))
        assert sleep_time > 0.01, "The value of sleep_time must be > 0.01, not {}".format(sleep_time)

        # Take 10 ms sleeps until sleep_time is reached
        # and check after each sleep if the Nimbus is still running
        elapsed_time = 0
        while elapsed_time <= sleep_time and self.running:
            if not self.running:
                break
            time.sleep(0.01)
            elapsed_time += 0.01
        return


import inspect
from pygame import image
from PIL import Image
import numpy as np


# Let there be ASCII art
logo = """                                        
 _____ _       _           _         _           
|   | |_|_____| |_ _ _ ___|_|___ ___| |_ ___ ___ 
| | | | |     | . | | |_ -| |   | .'|  _| . |  _|
|_|___|_|_|_|_|___|___|___|_|_|_|__,|_| |___|_|  
                                                  
                                               
RM Nimbus GUI for Python
                                            
"""


def message(text):
    """Debug message

    Prints a message in the console including the name of the function
    that sent the message.

    Args:
        text (str): The text of the message

    """

    # Get name of function that sent the message
    caller = inspect.stack()[1][3]
    # Print the debug message
    print('[nimbusinator] {}: {}'.format(caller, text))


def pil_to_pygame_image(PIL_image):
    """Convert PIL image to pygame surface

    Args:
        PIL_image (PIL.Image): The PIL image to be converted

    Returns:
        (pygame.image): The pygame image
    
    """

    return image.fromstring(PIL_image.tobytes(), PIL_image.size, PIL_image.mode)


def fix_coord(screen_size, coord):
    """Fix coordinates for use in OpenCV's drawing functions

    PIL images have upside-down co-ordinates and it shafts me
    every goddamn time, so this function "deals with it"

    Args:
        screen_size (tuple): The screen size (width, height)
        coord (tuple): The x, y coordinate to fix (x, y)

    Returns:
        (tuple): The coordinates flipped upside-down

    """
    
    return (coord[0], ((screen_size[1] - 1) - coord[1]))


def is_valid_colour(nimbus, colour):
    """Validate colour number

    If the colour number is out of range for the current screen
    mode a fatal error is yielded

    Args:
        nimbus (Nimbus): The Nimbus object
        colour (int): The colour number
    
    Returns:
        (bool): The test result

    """

    # validate low-res colour
    if nimbus.screen_mode == 'lo':
        if colour >= 0 and colour <= 15:
            # it's fine
            return True
        else:
            # it's not
            return False
    
    # validate high-res colour
    if nimbus.screen_mode == 'hi':
        if colour >= 0 and colour <= 3:
            # it's fine
            return True
        else:
            # it's not
            return False


def colrows_to_xy(screen_size, cursor_position):
    """Convert cursor position to x, y pixel position

    Args:
        screen_size (tuple): The screen size (width, height)
        cursor_position (tuple): The cursor position (row, col)

    Returns:
        (tuple): The screen position in pixels (x, y)
    
    """

    x = (8 * (cursor_position[0] - 1))
    y = (screen_size[1] - 2) - (cursor_position[1] * 10)
    return (x, y)


def ceildiv(a, b):
    return -(-a // b)


def char_image_selecta(font_img, ascii_code, font):
    """Get the image of a character from a PNG

    Enter an ASCII code and a transparent PNG image of the char is returned.
    Codes < 33 and delete char (127) just return a space.

    Args:
        ascii_code (int): The ASCII code (extended) of the character
        font (int): The font or charset

    Returns:
        (PIL image): The transparent image of the character
    
    """

    # On our Nimbus char map PNGs delete (127) is just a blank space, so if we
    # receive any < 33 control chars, set the ascii value to 127 so a space is
    # also returned in those cases.
    if ascii_code < 33:
        ascii_code = 127
    # Calculate the row and column position of the char on the character map PNG
    map_number = ascii_code - 32    # codes < 33 are not on the map (unprintable)
    row = ceildiv(map_number, 30)
    column = map_number - (30 * (row - 1))
    # Calculate corners of box around the char
    x1 = (column - 1) * 10
    x2 = x1 + 10
    y1 = (row - 1) * 10
    y2 = y1 + 10
    # Chop out the char and return
    return font_img.crop((x1, y1, x2, y2))


def recolour(nimbus, img, colour1, colour2, has_alpha=False):
    """Replace one colour with another in an image"""

    # Unpack the image
    img = img.convert('RGBA')
    data = np.array(img)
    if has_alpha:
        r, g, b, a = data.T
    else:
        r, g, b = data.T

    # Unpack colour1 and 2
    r1, g1, b1 = colour1

    # Replace white with red... (leaves alpha values alone...)
    white_areas = (r == r1) & (b == b1) & (g == g1)
    data[..., :-1][white_areas.T] = colour2

    return Image.fromarray(data)


def points_to_nparray(points_list):
    """Convert points_list from a set_points command into a numpy array"""

    empty = [255, 255, 255, 0]
    fill = [0, 0, 0, 255]

    rows = []

    for points in points_list:
        row = []
        for char in points:
            if char == '.':
                row.append(fill)
            if char == ' ':
                row.append(empty)
        rows.append(row)

    return np.array(rows, dtype=np.uint8)


from pynput import keyboard


supported_control_keys = {
                keyboard.Key.f1: '__F1__',
                keyboard.Key.f2: '__F2__',
                keyboard.Key.f3: '__F3__',
                keyboard.Key.f4: '__F4__',
                keyboard.Key.f5: '__F5__',
                keyboard.Key.f6: '__F6__',
                keyboard.Key.f7: '__F7__',
                keyboard.Key.f8: '__F8__',
                keyboard.Key.f9: '__F9__',
                keyboard.Key.f10: '__F10__',
                keyboard.Key.ctrl_l: '__CTRL_L__',
                keyboard.Key.ctrl_r: '__CTRL_R__',
                keyboard.Key.enter: '__ENTER__',
                keyboard.Key.backspace: '__BACKSPACE__',
                keyboard.Key.esc: '__ESC__',
                keyboard.Key.delete: '__DELETE__',
                keyboard.Key.home: '__HOME__',
                keyboard.Key.page_up: '__PAGE_UP__',
                keyboard.Key.end: '__END__',
                keyboard.Key.page_down: '__PAGE_DOWN__',
                keyboard.Key.insert: '__INSERT__',
                keyboard.Key.up: '__UP__',
                keyboard.Key.down: '__DOWN__',
                keyboard.Key.left: '__LEFT__',
                keyboard.Key.right: '__RIGHT__',
                keyboard.Key.tab: '__TAB__'
}


import setuptools

with open('readme.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name="nimbusinator",
    version="0.2.1",
    author="Tim Adams",
    author_email="adamstimb@gmail.com",
    description="RM Nimbus GUI for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/adamstimb/nimbusinator",
    packages=setuptools.find_packages(),
    package_data={'nimbusinator': ['data/*']},
    install_requires=[
        "PySDL2",
        "Pillow",
        "numpy",
        "simpleaudio",
        "pynput==1.4.5",
        "psutil",
        "pygame>=1.9.3,<2.0;python_version!='3.3'",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import platform
from psutil import virtual_memory

def welcome(cmd, nim):
    """RM Nimbus Welcome Screen

    This simulates the original Welcome Screen shown when the
    Nimbus is switched on and is waiting to boot from a disk.

    Your computer's memory size will be reported to keep things
    authentic.  Your Python version will substitute for
    Nimbus firmware version and the version number of your OS
    will be used as the machine serial number.

    After the Welcome Screen we imitate the old MS-DOS loading
    covfefe by printing the name of your OS and then the full
    version string of your Python distro.

    Oh, and turn up your speakers to hear the floppy drive :D

    """

    # Get Python version
    py_version = '{}.{}.{}'.format(
        platform.sys.version_info.major,
        platform.sys.version_info.minor, 
        platform.sys.version_info.micro
    )
    # Get memory stats
    mem = virtual_memory()
    total_mem = int(mem.total / 1024 / 1024)
    used_mem = int(mem.used / 1024 / 1024)
    free_mem = int(mem.free / 1024 / 1024)
    # Set up screen
    cmd.set_mode(80)
    cmd.set_colour(0, 0)
    cmd.set_colour(1, 9)
    cmd.set_border(1)
    # Frame
    cmd.area([(0, 0), (639, 0), (639, 249), (0, 249), (0, 0)], brush=2)
    cmd.area([(3, 2), (636, 2), (636, 247), (3, 247), (3, 2)], brush=1)
    # Nimbus logo with frame
    xl = 10
    yl = 212
    frame = [
        (xl, yl),
        (xl+304, yl),
        (xl+304, yl+32),
        (xl, yl+32),
        (xl, yl)
    ]
    cmd.plonk_logo((xl, yl))
    cmd.line(frame, brush=2)
    # Welcome
    cmd.plot('Welcome', (238, 145), size=3, brush=0, font=1)
    cmd.plot('Welcome', (236, 147), size=3, brush=2, font=1)
    # Memory
    cmd.plot('total memory size {m: >5} Mbytes'.format(m=total_mem), (15, 5), brush=0, font=1)
    cmd.plot('used  memory size {m: >5} Mbytes'.format(m=used_mem), (15, 15), brush=0, font=1)
    cmd.plot('main  memory size {m: >5} Mbytes'.format(m=free_mem), (15, 25), brush=0, font=1)
    # Version info
    cmd.area([(393, 4), (632, 4), (632, 30), (393, 30), (393, 4)], brush=2)
    cmd.area([(395, 5), (629, 5), (629, 29), (395, 29), (395, 5)], brush=3)
    cmd.plot('Firmware version: {}'.format(py_version), (400, 16), brush=0, font=1)
    cmd.plot('Serial number: {}'.format(platform.release()[:8]), (400, 6), brush=0, font=1)
    # Please supply an operating system
    cmd.plot('Please supply an operating system', (188, 100), brush=3, font=1)
    nim.sleep(1.3)
    cmd.plot('Please supply an operating system', (188, 100), brush=1, font=1)
    # Looking for an operating system
    cmd.plot('Looking for an operating system - please wait', (140, 100), brush=3, font=1)
    nim.run_floppy(True)
    nim.sleep(2)
    # Loading operating system
    cmd.plot('Looking for an operating system - please wait', (140, 100), brush=1, font=1)
    cmd.plot('Loading operating system', (224, 100), brush=3, font=1)
    nim.sleep(4)
    # Pretend DOS boot sequence
    cmd.set_mode(80)
    cmd.set_colour(0, 0)
    cmd.cls()
    cmd.set_cursor(True)
    nim.sleep(1.3)
    os_string = '{} - Version {}'.format(platform.system(), platform.release())
    os_string = os_string[:70]
    os_string = '   {}   '.format(os_string)
    # print os string in a groovy box thing
    bar = ''
    for i in range(0, len(os_string)):
        bar += ' '
    cmd.set_paper(3)
    cmd.set_pen(0)
    cmd.print(bar)
    cmd.print(os_string)
    cmd.print(bar)
    cmd.set_paper(0)
    cmd.set_pen(3)
    cmd.print(' ')
    nim.sleep(1.7)
    python_version = 'Python {}'.format(platform.sys.version)
    for text in python_version.split('\n'):
        cmd.print(text)
    nim.sleep(3.2)
    nim.run_floppy(False)
    nim.sleep(1.2)
    cmd.set_cursor(False)
    # Done
